﻿-- Handy shortcuts
local wv = BeneCast.WorkVars;
local BeneCastButton = BeneCast.Button;
local BeneCastPanel = BeneCast.Panel;
local BeneCastPanelDropDown = BeneCast.Panel.DropDown;
local hp = BeneCast.Helper;
local L = BeneCast.Localisation;
local const = BeneCast.Constants;

BeneCast.ToggleConfigWindow = function()

	local cd = LibStub("AceConfigDialog-3.0");
	if cd.OpenFrames['BeneCast'] then
		cd:Close('BeneCast');
	else
		cd:Open('BeneCast');
	end

end

-- Functions for wv.Setup
do
	BeneCast.GetSnapToOption = function(info)

		BeneCast:Debug('BeneCast.GetSnapToOption()');
		if wv.Setup.snapto == nil then
			wv.Setup.snapto = 'BeneCast';
		end
		return wv.Setup.snapto;
		
	end

	BeneCast.SetSnapToOption = function(info, value)

		BeneCast:Debug('BeneCast.SetSnapToOption()');
		wv.Setup.snapto = value;
		BeneCast.SnapTo:ApplyConfigTable();
		BeneCast.SnapTo:SnapPanels();
		
	end

	BeneCast.GenerateSnapToOptions = function(info)

		BeneCast:Debug('BeneCast.GenerateSnapToOptions()');
		local snaptooptions = hp.new();
		for addonname, addon in pairs(wv.SnapToFound) do
			snaptooptions[addonname] = addon.name;
		end
		return snaptooptions;
		
	end

	BeneCast.GetSetupOption = function(info)

		return wv.Setup[info[#info]];
		
	end

	BeneCast.SetSetupOption = function(info,value)

		local varname;
		varname = info[#info];
		wv.Setup[varname]=value;
		BeneCast:Print(varname .. " was set to: " .. tostring(value) );
		if varname == 'hideminimapbutton' then
			if wv.MinimapButton then
				if value then
					wv.MinimapButton:Hide();
				else
					wv.MinimapButton:Show();
				end
			end
			
		elseif varname == 'buttonsize' then
			BeneCast:ResizeButtons();
		
		elseif varname == 'flashasfade' or option == 'showallbuttons' then
			BeneCast:UpdateAllPanels();
			
		elseif varname == 'checkrange' then
			if value then
				BeneCast.Timers.CheckRange:Enable();
			else
				BeneCast.Timers.CheckRange:Disable();
			end
		
		elseif varname == 'showpets' then
			-- Attach the panels again, just to be sure
			BeneCast.SnapTo:SnapPanels();
			
			for pet, owner in pairs(const.OwnersByPet) do
				BeneCast:UpdatePanelsForUnit(pet);
			end
			
		elseif varname == 'showtargettarget' then
			-- Attach the panels again, just to be sure
			BeneCast.SnapTo:SnapPanels();
			
			-- Redraw the buttons for the target of target
			BeneCast:UpdatePanelsForUnit('targettarget');
		
		end

	end
end

-- Functions for wv.Setup.config[configname]
do
	BeneCast.GetActionSetting = function(info)
		local configname = info[1];
		local config = wv.Setup.config[configname];
		if config[info.arg] == nil then
			config[info.arg] = { enabled = false };
		end
		local setting = config[info.arg];
		return setting.enabled;
	end

	BeneCast.SetActionSetting = function(info, val)
		local configname = info[1];
		local mini, maxi;
		if wv.Setup.cascadeplayer and configname == 'player' then
			mini = 1;
			maxi = hp._CountMembers(const.Configs);
		else
			mini = hp._GetKeyForVal(const.Configs, configname);
			maxi = mini;
		end
		for i = mini, maxi do
			configname = const.Configs[i];
			local config = wv.Setup.config[configname];
			if config[info.arg] == nil then
				config[info.arg] = { enabled = false };
			end
			local setting = config[info.arg];
			setting.enabled = val;
		end
		BeneCast:UpdateAllPanels();
	end
end

-- Functions for wv.ButtonInfo, wv.EffectInfo, wv.CustomButtons and wv.CustomEffects
do
	-- Helper functions
	do
		function wv.GUIState:GenerateIndexTable(force)
			if self.indextable ~= nil then
				if not force then
					return;
				else
					hp._ClearTable(self.indextable);
				end
			else
				self.indextable = hp.new();
			end
			for name, _ in pairs(self.InfoTable) do
				tinsert(self.indextable, name);
			end
		end

		function wv.GUIState:GetIcon(i)
			local icon;
			if i ~= nil and self.indextable[i] and self.InfoTable[self.indextable[i]] then
				icon = self.InfoTable[self.indextable[i]].icon;
			end
			if icon == nil or icon == '' then
				icon = self.defaulticon;
			end
			return icon;
		end

		function wv.GUIState:InitCurrent()
			if self.name_in_edit then
				return;
			end
			if self.current == nil then
				if self.indextable[1] then
					self.current = 1;
					self.currenticon = self:GetIcon(self.current);
					self.previous = nil;
					self.previousicon = self:GetIcon(self.previous);
					if self.indextable[self.current + 1] then
						self.next = self.current + 1;
					else
						self.next = nil;
					end
					self.nexticon = self:GetIcon(self.next);
				end
			end
		end

		function wv.GUIState:UpdateCurrentItem()
			if self.current == nil or self.name_in_edit then
				return;
			end
			hp._ClearTable(self.currentitem);
			hp._CopyTableToExisting(self.InfoTable[self.indextable[self.current]], self.currentitem, true);
			self.currenticon = self:GetIcon(self.current);
			if self.current == 1 then
				self.previous = nil;
			else
				self.previous = self.current - 1;
			end
			self.previousicon = self:GetIcon(self.previous);
			if self.indextable[self.current + 1] then
				self.next = self.current + 1;
			else
				self.next = nil;
			end
			self.nexticon = self:GetIcon(self.next);
		end

		wv.GUIState.GetSelf = function(info)
			if info[2] == 'buttons' then
				return wv.CustomButtonGUIState;
			elseif info[2] == 'effects' then
				return wv.CustomEffectGUIState;
			else
				error('wv.GUIState.PreviousIcon(): info[2] = '..info[2]);
			end
		end
	end
	
	-- Control functions
	do
		function wv.GUIState:Previous(info)
			if self.name_in_edit then
				return;
			end
			self:GenerateIndexTable();
			self:InitCurrent();
			if self.current ~= nil and self.current ~= 1 then
				self.current = self.current - 1;
				self:UpdateCurrentItem();
			end
		end

		function wv.GUIState:PreviousConfirm(info)
			if self.name_in_edit then
				return;
			end
			if self.currentitem.changed then
				return 'Changes not saved, Continue?';
			else
				return nil;
			end
		end

		function wv.GUIState:PreviousDisabled(info)
			self:GenerateIndexTable();
			self:InitCurrent();
			return not (self.current ~= nil and self.current ~= 1) or self.in_name_edit;
		end

		wv.GUIState.PreviousIcon = function(info)
			return wv.GUIState.GetSelf(info).previousicon;
		end

		function wv.GUIState:Next(info)
			if self.in_name_edit then
				return;
			end
			self:GenerateIndexTable();
			self:InitCurrent();
			if self.current ~= nil and self.indextable[self.current + 1] then
				self.current = self.current + 1;
				self:UpdateCurrentItem();
			end
		end

		wv.GUIState.NextConfirm = wv.GUIState.PreviousConfirm;

		function wv.GUIState:NextDisabled(info)
			self:GenerateIndexTable();
			self:InitCurrent();
			return not (self.current ~= nil and self.indextable[self.current + 1]) or self.in_name_edit;
		end

		wv.GUIState.NextIcon = function(info)
			return wv.GUIState.GetSelf(info).nexticon;
		end

		wv.GUIState.CurrentIcon = function(info)
			return wv.GUIState.GetSelf(info).currenticon;
		end

		function wv.GUIState:CopyToNew(info)
			if self.name_in_edit or self.current == nil or self.currentitem.changed then
				return;
			end
			-- after copying, everything but 'Name' and 'Cancel Edit' is disabled until a valid name is entered
			self.currentitem.name = '';
			self.name_in_edit = 'newfromcurrent';
		end

		function wv.GUIState:CopyToNewDisabled(info)
			self:GenerateIndexTable();
			self:InitCurrent();
			return (self.name_in_edit or self.current == nil or self.currentitem.changed);
		end

		function wv.GUIState:NameGet(info)
			return self.currentitem.name;
		end

		function wv.GUIState:NameValidate(info, value)
			if value == nil then
				return 'nil names are not accepted';
			end
			if self.InfoTable[value] then
				return 'That name is already in use';
			end
			if value == '' then
				return 'Empty names are not accepted';
			end
			if not self.name_in_edit then
				return 'Name should not be editable';
			end
		end

		function wv.GUIState:NameSet(info, value)
			-- I assume it passed the validate
			if self.name_in_edit == 'newfromcurrent' then
				if self.currentitem.origin ~= 'custom' then
					self.currentitem.oldorigin = self.currentitem.origin;
				end
				self.currentitem.origin = 'custom';
			end
			self.currentitem.changed = true;
			self.name_in_edit = nil;
			self.currentitem.name = value;
		end

		function wv.GUIState:NameDisabled(info)
			return not self.name_in_edit;
		end

		function wv.GUIState:CancelEdit(info)
			self.name_in_edit = nil;
			-- Reload the item from the info-table
			self:UpdateCurrentItem();
		end

		function wv.GUIState:CancelEditDisabled(info)
			return not self.name_in_edit;
		end

		function wv.GUIState:OriginValues(info)
			return { spell = 'Spell', item = 'Item', custom = 'Custom', preload = 'Preloaded' };
		end

		function wv.GUIState:OriginGet(info)
			return self.currentitem.origin;
		end

		function wv.GUIState:OriginDisabled(info, value)
			return self.name_in_edit;
		end

		function wv.GUIState:OriginSet(info, value)
			-- Do nothing
		end

		function wv.GUIState:TargetValues(info)
			local t = hp.new();
			for code, v in pairs(const.TargetTypes) do
				t[code] = v.trans;
			end
			return t;
		end

		function wv.GUIState:TargetGet(info)
			return self.currentitem.target;
		end

		function wv.GUIState:TargetSet(info, value)
			if self.name_in_edit then
				return;
			end
			self.currentitem.target = value;
			if self.currentitem.origin ~= 'custom' then
				self.currentitem.oldorigin = self.currentitem.origin;
			end
			self.currentitem.origin = 'custom';
			self.currentitem.changed = true;
		end

		function wv.GUIState:TargetDisabled(info)
			return self.name_in_edit;
		end

		function wv.GUIState:AttributesGet(info)
			local ret = '';
			for k, v in pairs(self.currentitem.attribs) do
				if ret == '' then
					ret = k..' = '..v;
				else
					ret = ret..'\n'..k..' = '..v;
				end
			end
			return ret;
		end

		function wv.GUIState:AttributesSet(info, value)
			if self.name_in_edit then
				return;
			end
			hp._ClearTable(self.currentitem.attribs);
			local tbl = { strsplit('\n', value) };
			local type, val;
			for i, str in pairs(tbl) do
				type, val = strsplit('=', str);
				type = strtrim(type);
				val = strtrim(val);
				self.currentitem.attribs[type] = val;
			end
			if self.currentitem.origin ~= 'custom' then
				self.currentitem.oldorigin = self.currentitem.origin;
			end
			self.currentitem.origin = 'custom';
			self.currentitem.changed = true;
		end

		function wv.GUIState:AttributesDisabled(info)
			return self.name_in_edit;
		end

		function wv.GUIState:IconpathGet(info)
			return self.currentitem.icon;
		end

		function wv.GUIState:IconpathSet(info, value)
			-- I assume it passed the validate
			if self.name_in_edit then
				return;
			end
			self.currentitem.icon = value;
			self.currentitem.changed = true;
			if self.currentitem.origin ~= 'custom' then
				self.currentitem.oldorigin = self.currentitem.origin;
			end
			self.currentitem.origin = 'custom';
		end

		function wv.GUIState:IconpathDisabled(info)
			return self.name_in_edit;
		end

		wv.GUIState.IconImage = function(info)
			return wv.GUIState.GetSelf(info).currentitem.icon;
		end

		function wv.GUIState:TemplateValues(info)
			local t = hp.new();
			for code, v in pairs(BeneCast.ButtonTemplates) do
				t[code] = v.name;
			end
			return t;
		end

		function wv.GUIState:TemplateGet(info)
			return self.template;
		end

		function wv.GUIState:TemplateSet(info, value)
			if self.name_in_edit then
				return;
			end
			self.template = value;
		end

		function wv.GUIState:TemplateDisabled(info)
			return self.name_in_edit;
		end

		function wv.GUIState:CopyTemplateFunc(info)
			if self.name_in_edit or self.template == nil then
				return;
			end
			local templ = BeneCast.ButtonTemplates[self.template];
			if templ then
				hp._CopyTableToExisting(templ, self.currentitem, true);
			end
		end

		function wv.GUIState:CopyTemplateDisabled(info)
			return self.name_in_edit;
		end
	end
	
	hp._CopyTableToExisting(wv.GUIState, wv.CustomButtonGUIState, true);
	hp._CopyTableToExisting(wv.GUIState, wv.CustomEffectGUIState, true);
end

wv.DefaultOptions = {
    name = 'benecast',
    desc = const.Name .. ' ' .. const.Version,
    type = 'group',
    args = {
		win = {
			type = 'execute',
			name = 'win',
			desc = L['Settings'],
			guiHidden = true, -- The GUI doesn't need to show this
			func = BeneCast.ToggleConfigWindow,
			order = 1,
		},
		showvar = {
			type = 'execute',
			name = 'showvar',
			guiHidden = true,
			func = function()
				local rootname = BeneCast.MindReader:Build(BeneCast);
				BeneCast.MindReader:ShowBuild(rootname);
			end,
			order = 2,
		},
		general = {
			type = 'group',
			name = 'general',
			desc = L['General'],
			order = 3,
			args = {
				snapto = {
					type = 'select',
					name = 'snapto',
					desc = L['Snap To Options'],
					values = BeneCast.GenerateSnapToOptions,
					confirm = true,
					confirmText = 'This will reset the panel-positions, continue?',
					get = BeneCast.GetSnapToOption,
					set = BeneCast.SetSnapToOption,
					order = 1,
				},
				debug = {
					type = 'toggle',
					name = 'debug',
					desc = 'Debug',
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 2,
				},
				hideminimapbutton = {
					type = 'toggle',
					name = 'hideminimapbutton',
					desc = L['Hide the minimap button'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 3,
				},
				cascadeprofiles = {
					type = 'toggle',
					name = 'cascadeprofiles',
					desc = 'Cascade Profiles',
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 4,
				},
				orientation = {
					type = 'select',
					name = L['Orientation'],
					values = function(info)
						local vals = hp.new();
						local currentorient = 'RIGHT';
						local orientobj;
						while hp._CountMembers(vals) == 0 and currentorient ~= 'RIGHT' do
							orientobj = const.Orientations[currentorient];
							vals[currentorient] = orientobj.translation;
							currentorient = orientobj.next;
						end
						return vals;
					end,
					get = function(info)
						return wv.Setup.orientation;
					end,
					set = function(info,val)
						wv.Setup.orientation = val;
					end,
					order = 5,
				},
				buttonsize = {
					type = 'range',
					name = L['Button Size'],
					min = 8,
					max = 36,
					step = 1,
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 6,
				},
				flashasfade = {
					type = 'toggle',
					name = L['Flash as Fade'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 7,
				},
				checkrange = {
					type = 'toggle',
					name = L['Check Range for Target'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 8,
				},
				showtooltips = {
					type = 'toggle',
					name = L['Show Button Tooltips'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 9,
				},
				showtooltipname = {
					type = 'toggle',
					name = L['Spell Name Only'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 10,
				},
				showallbuttons = {
					type = 'toggle',
					name = L['Always Show Buttons'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 11,
				},
				unlockbuttons = {
					type = 'toggle',
					name = L['Unlock Buttons'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 12,
				},
				cascadeplayer = {
					type = 'toggle',
					name = L['Cascade Player Config'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 13,
				},
				showpets = {
					type = 'toggle',
					name = L['Show Pet Buttons'],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 14,
				},
				showtargettarget = {
					type = 'toggle',
					name = L["Show Target's Target"],
					get = BeneCast.GetSetupOption,
					set = BeneCast.SetSetupOption,
					order = 15,
				},
			},
		},
		player = {
			type = 'group',
			name = L['Player'],
			order = 4,
			args = { }, -- This will be filled as the spells load
		},
		pet = {
			type = 'group',
			name = L['Pet'],
			order = 5,
			args = { }, -- This will be filled as the spells load
		},
		party = {
			type = 'group',
			name = L['Group Members'],
			order = 6,
			args = { }, -- This will be filled as the spells load
		},
		partypet = {
			type = 'group',
			name = L['Group Pets'],
			order = 7,
			args = { }, -- This will be filled as the spells load
		},
		other = {
			type = 'group',
			name = L['Non-Group Members'],
			order = 8,
			args = { }, -- This will be filled as the spells load
		},
        custom = {
			type = 'group',
			name = 'custom',
			desc = L['Custom'],
			hidden = true, -- TODO: Finish this so it can become cmdHidden
			order = 9,
			args = {
				buttons = {
					type = 'group',
					name = L['Buttons'],
					handler = wv.CustomButtonGUIState,
					order = 1,
					args = {
						previous = {
							name = 'Previous',
							type = 'execute',
							func = 'Previous',
							disabled = 'PreviousDisabled',
							confirm = 'PreviousConfirm',
						},
						previcon = {
							name = 'PreviousIcon',
							type = 'description',
							image = wv.GUIState.PreviousIcon,
						},
						next = {
							name = 'Next',
							type = 'execute',
							func = 'Next',
							disabled = 'NextDisabled',
							confirm = 'NextConfirm',
						},
						nexticon = {
							name = 'NextIcon',
							type = 'description',
							image = wv.GUIState.NextIcon,
						},
						curricon = {
							name = 'CurrIcon',
							type = 'description',
							image = wv.GUIState.CurrentIcon,
						},
						copytonew = {
							name = 'Copy to New',
							type = 'execute',
							func = 'CopyToNew',
							disabled = 'CopyToNewDisabled',
						},
						name = { -- Usually non-editable unless copying to new button
							name = 'Name',
							type = 'input',
							get = 'NameGet',
							set = 'NameSet',
							validate = 'NameValidate',
							disabled = 'NameDisabled',
						},
						canceledit = { -- Usually non-editable unless name is in edit
							name = 'Cancel Edit',
							type = 'execute',
							func = 'CancelEdit',
							disabled = 'CancelEditDisabled',
						},
						origin = {
							name = 'Origin',
							desc = 'You cannot change this value',
							type = 'select',
							values = 'OriginValues',
							get = 'OriginGet',
							set = 'OriginSet',
							disabled = 'OriginDisabled',
						},
						target = {
							name = 'Target',
							type = 'select',
							values = 'TargetValues',
							get = 'TargetGet',
							set = 'TargetSet',
							disabled = 'TargetDisabled',
						},
						attribs = {
							name = 'Attributes',
							type = 'input',
							multiline = 10,
							get = 'AttributesGet',
							set = 'AttributesSet',
							disabled = 'AttributesDisabled',
						},
						iconpath = {
							name = 'Icon path',
							type = 'input',
							get = 'IconpathGet',
							set = 'IconpathSet',
							disabled = 'IconpathDisabled',
						},
						icon = {
							name = 'Icon',
							type = 'description',
							image = wv.GUIState.IconImage,
						},
						template = {
							name = 'Template',
							type = 'select',
							values = 'TemplateValues',
							get = 'TemplateGet',
							set = 'TemplateSet',
							disabled = 'TemplateDisabled',
						},
						copytemplate = {
							name = 'Copy From Template',
							type = 'execute',
							func = 'CopyTemplateFunc',
							disabled = 'CopyTemplateDisabled',
						},
						--[[
						TODO:
						delete = {
							type = 'execute',
						},
						save = {
							type = 'execute',
						},
						cancel = {
							type = 'execute',
						},
						--]]
					},
				},
				effects = {
					type = 'group',
					name = L['Effects'],
					order = 1,
					handler = wv.CustomEffectGUIState,
					args = {
						-- TODO: copy from buttons and add overrides and enables, minus unneeded stuff
					},
				},
			},
        },
    },
};

-- Function to generate a config for a certain button
-- For now this is a toggle, but later it can be anything
function BeneCast:ActionToConfig(name, group)

	return {
		type = 'toggle',
		name = name,
		arg = group,
		get = BeneCast.GetActionSetting,
		set = BeneCast.SetActionSetting,
	};

end

-- Function to generate entire spellconfig by the known button-list
function BeneCast:GenerateConfigFor(configtable, configname)

	local propername, count;
	count = 0;
	
	for group, button in pairs(wv.ButtonInfo) do
		if button.target then
			if const.TargetTypes[button.target][configname] then
				if button.origin ~= 'custom' or (button.origin == 'custom' and wv.CustomButtons[group].publish) then
					propername = string.gsub(button.name,' ','');
					configtable[propername] = self:ActionToConfig(button.name, group);
					count = count + 1;
				end
			end
		else
			self:Debug('button '..button.name..' has no target-node');
		end
	end
	
	return count;
	
end

-- Function to do the above for ALL configs
function BeneCast:GenerateAllConfigs(optiontable)

	local configtable, count;
	for i, configname in pairs(const.Configs) do
		configtable = optiontable.args[configname].args;
		count = self:GenerateConfigFor(configtable, configname);
		--self:Debug('Added '..count..' actions to config='..configname);
	end

end